import pathlib

import pytest

from pimht import (
    __version__,
    from_bytes,
    from_string,
    from_filename,
    from_fileobj,
    MHTML,
    MHTMLPart,
)


EXAMPLE_PATH = filepath = pathlib.Path(__file__).parent / "example_com.mhtml"
EXAMPLE_URL = "https://www.example.com/"


def test_parsers():
    with open(EXAMPLE_PATH, "rb") as fp:
        bytes_data = fp.read()

    bytes_mhtml = from_bytes(bytes_data)
    assert isinstance(bytes_mhtml, MHTML)

    str_mhtml = from_string(bytes_data.decode("utf-8"))
    assert isinstance(str_mhtml, MHTML)

    filename_mhtml = from_filename(EXAMPLE_PATH)
    assert isinstance(filename_mhtml, MHTML)

    with open(EXAMPLE_PATH, "r", encoding="utf-8") as fp:
        str_file_mhtml = from_fileobj(fp)
        assert isinstance(str_file_mhtml, MHTML)

    with open(EXAMPLE_PATH, "rb") as fp:
        bytes_file_mhtml = from_fileobj(fp)
        assert isinstance(bytes_file_mhtml, MHTML)


def test_parsing():
    mhtml = from_filename(EXAMPLE_PATH)

    assert mhtml.headers.get("From") == "<Saved by Blink>"
    assert mhtml.headers.get("Snapshot-Content-Location") == EXAMPLE_URL

    counter = 0
    for counter, part in enumerate(mhtml, start=1):
        assert isinstance(part, MHTMLPart)
        assert isinstance(part.headers, dict)

        if counter == 1:
            assert part.headers.get("Content-Location") == EXAMPLE_URL
            assert part.content_type == "text/html"
            assert part.is_text is True
            assert isinstance(part.raw, bytes)
            assert part.raw.startswith(b"<!DOCTYPE html><html>")
            assert part.text.startswith("<!DOCTYPE html><html>")

        elif counter == 2:
            assert part.content_type == "text/css"
            assert part.is_text is True
            assert isinstance(part.raw, bytes)
            assert part.raw.startswith(b'@charset "utf-8";')
            assert part.text.startswith('@charset "utf-8";')

        elif counter == 3:
            assert part.content_type == "image/png"
            assert part.is_text is False
            assert isinstance(part.raw, bytes)
            assert part.raw.startswith(b'\x89PNG')

    assert counter == 3


def test_parts_property():
    mhtml = from_filename(EXAMPLE_PATH)
    assert len(mhtml.parts) == 3
    # .parts and __iter__ return the same objects
    assert list(mhtml) is not mhtml.parts
    for a, b in zip(mhtml, mhtml.parts):
        assert a is b


def test_charset():
    mhtml = from_filename(EXAMPLE_PATH)
    html_part = mhtml.parts[0]
    # example_com.mhtml has charset in Content-Type via quoted-printable encoding
    # CSS part may or may not have charset
    assert html_part.charset is None or isinstance(html_part.charset, str)


def test_round_trip():
    mhtml = from_filename(EXAMPLE_PATH)
    data = mhtml.to_bytes()
    mhtml2 = from_bytes(data)

    assert mhtml.headers == mhtml2.headers
    assert len(mhtml.parts) == len(mhtml2.parts)

    for orig, rt in zip(mhtml.parts, mhtml2.parts):
        assert orig.headers == rt.headers
        assert orig.raw == rt.raw
        if orig.is_text:
            assert orig.text == rt.text


def test_set_text():
    mhtml = from_filename(EXAMPLE_PATH)
    part = mhtml.parts[0]
    original = part.text

    assert "Example Domain" in original
    part.text = original.replace("Example Domain", "Modified Domain")

    assert "Modified Domain" in part.text
    assert "Example Domain" not in part.text

    # round-trip preserves the modification
    mhtml2 = from_bytes(mhtml.to_bytes())
    assert "Modified Domain" in mhtml2.parts[0].text
    assert "Example Domain" not in mhtml2.parts[0].text

    # unmodified parts are identical
    for i in range(1, len(mhtml.parts)):
        assert mhtml.parts[i].raw == mhtml2.parts[i].raw


def test_set_raw():
    mhtml = from_filename(EXAMPLE_PATH)
    png_part = mhtml.parts[2]
    assert png_part.content_type == "image/png"

    png_part.raw = b"\x89PNG\r\n\x1a\nFAKE"
    assert png_part.raw == b"\x89PNG\r\n\x1a\nFAKE"

    # round-trip preserves the modification
    mhtml2 = from_bytes(mhtml.to_bytes())
    assert mhtml2.parts[2].raw == b"\x89PNG\r\n\x1a\nFAKE"


def test_set_text_invalidates_raw_cache():
    mhtml = from_filename(EXAMPLE_PATH)
    part = mhtml.parts[0]

    old_raw = part.raw
    part.text = "completely new content"
    assert part.raw != old_raw
    assert part.text == "completely new content"


def test_set_raw_invalidates_text_cache():
    mhtml = from_filename(EXAMPLE_PATH)
    part = mhtml.parts[0]

    old_text = part.text
    part.raw = b"new raw content"
    assert part.text != old_text
    assert part.raw == b"new raw content"


def test_raw_is_cached():
    mhtml = from_filename(EXAMPLE_PATH)
    part = mhtml.parts[0]
    assert part.raw is part.raw


def test_text_not_text_raises():
    mhtml = from_filename(EXAMPLE_PATH)
    png_part = mhtml.parts[2]
    assert png_part.is_text is False
    with pytest.raises(TypeError):
        _ = png_part.text



import re


BOUND_RE = re.compile(r";\s*boundary=(['\"])([^'\"]+)\1")


def parse_headers(raw: bytes) -> dict[str, str]:
    """
    Parse MIME headers from raw bytes into a str dict.
    Supports tab/space continuation of previous header line.
    """
    headers = {}

    key = None
    for line in raw.decode("ascii").split("\n"):
        if not line:
            continue

        # continuation
        if line[0] in (" ", "\t"):
            if key:
                headers[key] += line[1:]
            continue

        key, value = line.split(":", 1)
        headers[key.strip()] = value.strip()

    return headers


def find_boundary(content_type: str) -> bytes:
    """Extract and return the MIME boundary as bytes."""
    if match := BOUND_RE.search(content_type):
        return b"--" + match.group(2).encode("ascii")

    raise TypeError("No boundary in Content-Type")



import functools
import importlib.metadata
import quopri
import typing

from . import util

try:
    import pybase64 as base64
except ModuleNotFoundError:
    import base64

try:
    import cchardet as chardet
except ModuleNotFoundError:
    import chardet


__version__ = importlib.metadata.version("pimht")


class MHTMLPart:
    """Part of an MHTML archive."""

    def __init__(self, headers: dict, content: bytes):
        self.headers = headers
        self.content = content

    def __setattr__(self, name, value):
        if name == "raw":
            encoding = self.headers.get("Content-Transfer-Encoding")
            if encoding == "base64":
                self.content = base64.encodebytes(value).rstrip(b"\n")
            elif encoding == "quoted-printable":
                self.content = quopri.encodestring(value).rstrip(b"\n")
            else:
                raise ValueError(f"Unknown mhtml part encoding: {encoding}")

            self.__dict__["raw"] = value
            self.__dict__.pop("text", None)

        elif name == "text":
            charset = self.charset or "utf-8"
            self.raw = value.encode(charset)
            self.__dict__["text"] = value

        else:
            super().__setattr__(name, value)

    @functools.cached_property
    def content_type(self) -> str:
        """Content-type of the MHTML part."""
        return self.headers.get("Content-Type").split(";")[0]

    @functools.cached_property
    def is_text(self) -> bool:
        """Whether or not the parts content-type begins with 'text'."""
        return self.content_type.startswith("text/")

    @functools.cached_property
    def charset(self):
        """Charset from Content-Type header, or None."""
        ct = self.headers.get("Content-Type", "")
        for param in ct.split(";")[1:]:
            param = param.strip()
            if param.lower().startswith("charset="):
                return param.split("=", 1)[1].strip().strip("'\"")
        return None

    @functools.cached_property
    def raw(self) -> bytes:
        """The raw (bytes) content of the MHTML part."""
        encoding = self.headers.get("Content-Transfer-Encoding")
        if encoding == "base64":
            return base64.b64decode(self.content)
        if encoding == "quoted-printable":
            return quopri.decodestring(self.content)
        raise ValueError(f"Unknown mhtml part encoding: {encoding}")

    @functools.cached_property
    def text(self) -> str:
        """
        String representation of the MHTML part.
        Raises TypeError if the part cannot be decoded.
        """
        if not self.is_text:
            raise TypeError("MHTMLPart is not text")

        charset = self.charset
        if charset:
            return self.raw.decode(charset, "ignore")

        detected = chardet.detect(self.raw)
        return self.raw.decode(detected["encoding"] or "utf-8", "ignore")

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} headers={self.headers}>"


class MHTML:
    """
    MHTML archive.
    Objects are iterable to allow processing/inspection of archive contents.
    """

    def __init__(self, content: bytes):
        # Normalize line endings (matches text-mode behavior, simplifies parsing)
        content = content.replace(b"\r\n", b"\n")

        sep = content.find(b"\n\n")
        self.headers = util.parse_headers(content[:sep])
        self.boundary = util.find_boundary(self.headers["Content-Type"])
        self._body = content[sep + 2:]

    @functools.cached_property
    def parts(self) -> list:
        """All parts in the MHTML archive."""
        result = []
        segments = self._body.split(self.boundary)
        for segment in segments[1:]:
            if segment[:2] == b"--":
                break

            sep = segment.find(b"\n\n")
            if sep == -1:
                continue

            headers = util.parse_headers(segment[1:sep])
            content = segment[sep + 2:]
            if content.endswith(b"\n"):
                content = content[:-1]

            result.append(MHTMLPart(headers, content))
        return result

    def __iter__(self) -> typing.Iterator[MHTMLPart]:
        return iter(self.parts)

    def to_bytes(self) -> bytes:
        """Serialize the MHTML archive to bytes."""
        chunks = []

        # Global headers
        for key, value in self.headers.items():
            chunks.append(f"{key}: {value}\r\n".encode("ascii"))
        chunks.append(b"\r\n")

        # Parts
        for part in self.parts:
            chunks.append(self.boundary + b"\r\n")
            for key, value in part.headers.items():
                chunks.append(f"{key}: {value}\r\n".encode("ascii"))
            chunks.append(b"\r\n")
            chunks.append(part.content.replace(b"\n", b"\r\n"))
            chunks.append(b"\r\n")

        # Closing boundary
        chunks.append(self.boundary + b"--\r\n")

        return b"".join(chunks)

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} headers={self.headers}>"


def from_bytes(mhtml_bytes: bytes) -> MHTML:
    """Parse bytes as an MHTML object."""
    return MHTML(mhtml_bytes)


def from_string(mhtml_str: str) -> MHTML:
    """Parse a string as an MHTML object."""
    return MHTML(mhtml_str.encode("ascii"))


def from_fileobj(fileobj: typing.IO) -> MHTML:
    """Parse a file object as an MHTML object."""
    data = fileobj.read()
    if isinstance(data, str):
        data = data.encode("ascii")
    return MHTML(data)


def from_filename(filename: str) -> MHTML:
    """Parse the contents of a file path as an MHTML object."""
    with open(filename, "rb") as f:
        return MHTML(f.read())
